@@ -75,7 +75,7 @@ from types import GeneratorType, ModuleType
 from collections import deque
 from binascii import a2b_hex, b2a_hex
 
-VERSION = '0.05'
+VERSION = '0.06'
 
 PYTHON3 = sys.version_info >= (3, 0)
 if PYTHON3:
@@ -466,6 +466,7 @@ class Connection:
         self._recv_expect_hmac = None
         self._recv_hmac = None
         self._send_hmac = None
+        self._peer_version = None
         self._init_data = {'version': VERSION}
         
         self._my_nonce = struct.pack('>d', time.time()) + os.urandom(HMAC_SIZE - 8)
@@ -567,6 +568,7 @@ class Connection:
         return True
 
     def _got_init(self, p):
+        self._peer_version = p.get('version')
         return True
 
     def _got_keepalive(self):
@@ -736,10 +738,13 @@ class Client(Connection):
         self._garbage_proxies = []
 
         self._is_ready = False
-
         self._defmod = definitions_module(definitions)
         self._init_data['definitions_module'] = definitions
 
+        ## Compatibility with 0.05 server
+        self._init_data['definitions_source'] = []
+        self._peer_version = None
+
         if hasattr(sock_or_host, 'send'):
             self.set_socket(sock_or_host)
         elif sock_or_host is not None:
@@ -751,14 +756,21 @@ class Client(Connection):
     def _after_connect(self):
         self._is_ready = False
         self._pending_requests = []
+        self._init_response = None
 
-        source = self._defmod._source_code
-        self._definition_count = len(source)
-        self._init_data['definitions_source'] = source
+        self._definition_count = 0
 
         self._sock.setblocking(0)
         self._init_thread()
         self.send_challenge()
+        if self._defmod._source_code:
+            self.update_definitions(oncomplete=lambda val: None, 
+                                    onerror=self._init_error)
+
+    def _init_error(self, exc):
+        self.report_exception(exc)
+        ## Raise the error in all outstanding requests
+        self.close(exc)
 
     def read_response(self, block=True):
         '''Try to read a response and call its completion function (if any).
@@ -818,14 +830,13 @@ class Client(Connection):
             return True
 
         # index 0 = oncomplete, index 1 = onerror
-        compf = complete[int(type(p) is RemoteException)]
+        compf = complete[int(type(response) is RemoteException)]
         if isinstance(compf, tuple):
             compf[0](response, *compf[1:])
         else:
             compf(response)
         return True
 
-
     def send_request(self, oid, meth, args, kwargs={}, nonblocking=False,
                      oncomplete=None, onerror=None, threadid=None):
         '''Send a request to the remote server. This method is normally not
@@ -923,14 +934,17 @@ class Client(Connection):
         while not nbr._have_response:
             self.read_response()
 
-    def update_definitions(self):
+    def update_definitions(self, nonblocking=False,
+                           oncomplete=None, onerror=None, threadid=None):
         '''Call this after calling define_common to send any new common
         definitions to the server.'''
 
         source = self._defmod._source_code
         current_count = self._definition_count
-        self.send_request(0, 'update_definitions', (source[current_count:],), {})
+        rv = self.send_request(0, 'update_definitions', (source[current_count:],), {}, 
+                          nonblocking, oncomplete, onerror, threadid)
         self._definition_count = len(source)
+        return rv
 
     def raw_call(self, func, args, kwargs={}, nonblocking=False,
                         oncomplete=None, onerror=None, threadid=None):
@@ -1000,6 +1014,11 @@ class Client(Connection):
         return self.send_request(0, 'import_modules', ([name],), {},
                                  nonblocking, oncomplete, onerror, threadid)
 
+    @property
+    def server_version(self):
+        '''Return the version of EasyCluster on the server'''
+        return self._peer_version
+
     def __getattr__(self, attr):
         '''Returns a proxy object to call a global function or class on the
         server. This lets you do things like:
@@ -1066,7 +1085,9 @@ class Client(Connection):
         return prox
 
     def close(self, message='Connection closed'):
-        '''Close the connection and invalidate all remote proxies.'''
+        '''Close the connection and invalidate all remote proxies. `message` is
+        either a string or a RemoteException instance to raise in all
+        outstanding requests.'''
         Connection.close(self)
         proxies = self._proxy_by_id
         while proxies:
@@ -1074,9 +1095,13 @@ class Client(Connection):
             prox = wrprox()
             if prox is not None:
                 prox._connection = _dummy_connection
-
+                
         ## Send an error to any outstanding requests
-        error = RemoteException(IOError(message), '')
+        if isinstance(message, RemoteException):
+            error = message
+        else:
+            error = RemoteException(IOError(message), '')
+
         comp = self._completion
         while comp:
             rid, (oncomplete, onerr) = comp.popitem()
@@ -1379,8 +1404,8 @@ class Server(Connection):
 
 
     def _got_init(self, args):
+        super(Server, self)._got_init(args)
         self._root._defmod = definitions_module(args['definitions_module'], False)
-        self._root.update_definitions(args['definitions_source'])
         self._keep_alive_interval = args.get('keep_alive_interval')
         if self._keep_alive_interval:
             self._sock.setblocking(0)
@@ -1496,21 +1521,30 @@ class ClientGroup:
         '''Read responses from the Client objects in the group.'''
         etime = (None if timeout is None else time.time() + timeout)
         while max:
-            rtime = (1000.0 if etime is None else max(0, etime - time.time()))
+            rtime = (1000.0 if etime is None else etime - time.time())
+            if rtime < 0:
+                rtime = 0
             if self.epoll:
                 evts = self.epoll.poll(rtime, len(self.clients))
                 if not evts and not rtime:
                     return
                 for fd, evt in evts:
                     if self.clients[fd].read_response(False):
-                        max -= 1
+                        if max > 0:
+                            max -= 1
+                            if max == 0:
+                                return
+                        
             else:
                 r, _, _ = select.select(self.clients.values(), (), (), rtime)
                 if not r and not rtime:
                     return
                 for client in r:
                     if client.read_response(False):
-                        max -= 1
+                        if max > 0:
+                            max -= 1
+                            if max == 0:
+                                return
 
 class QuietServer(Server):
     '''A subclass of Server which does not print anything to output.'''
@@ -1542,7 +1576,7 @@ def definitions_module(name, keep_source=True):
     def define(code):
         if not code.endswith('\n'):
             code += '\n'
-        cb = compile(code, '<easycluster definitions %r>' % name, 'exec')
+        cb = compile(code, '<easycluster definitions %r>' % name, 'exec', dont_inherit=True)
         eval(cb, mod.__dict__)
         if keep_source:
             source.append(code)
